'use strict'

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const lodash = require('lodash')
const qe = require('./query-expressions')
const actionCache = require('./action-cache')
const async = require('async')
const httpUtils = require('./http-api/http-utils')

const DEFAULT_INTERVAL = 5000

exports.create = create

function create (context) {
  return new ActionManager(context)
}

class ActionManager {
  constructor (context) {
    this._context = context
    this._savedViews = new Map()
    this._notifications = new Map()
    this._actions = require('./actions')(context)
    this._actionCache = actionCache.create()

    this.interval = this._context.config.view.interval || DEFAULT_INTERVAL

    // prune views that are gone
    // update edited views
    // add new views
    this._context.eventManager.on('savedViews-settings-changed', () => {
      this.updateActionsList()
      this.updateIntegrations()
    })

    this._context.eventManager.on('integrations-settings-changed', () => {
      this.updateIntegrations()
    })

    this._context.eventManager.on('notifications-settings-changed', () => {
      this.updateNotificationsList()
      this.updateIntegrations()
    })

    this._context.eventManager.on('agent-found', evt => {
      const agent = this.getAgent(evt)
      if (agent) this.sendAgentFound(agent)
    })

    this._context.eventManager.on('agent-exit', evt => {
      const agent = this.getAgent(evt)
      if (!agent) return
      this.sendAgentExit(agent, lodash.pick(evt.args, ['exitCode', 'error']), evt.args.profileId)

      // Since the process is now dead, delete _blocked cache entry for this
      // process if one exists, and send a notification.  This lookup should
      // succeed despite not having the loop_id and callback_cntr because of
      // the trie lookup.
      const blockedEvent = this._actionCache._blocked.get(evt.agent)
      if (blockedEvent && blockedEvent.type === 'blocked') {
        this._actionCache._blocked.delete(evt.agent)
        this.sendProcessBlocked(agent, {
          stack: blockedEvent.event.args.stack,
          threshold: blockedEvent.event.blockedAt ? (Date.now() - blockedEvent.event.blockedAt) : blockedEvent.event.args.threshold
        })
      }
    })

    // This fires at the threshold, but the process is still blocked.
    // Wait until we get process-unblocked
    this._context.eventManager.on('process-blocked', evt => {
      evt.blockedAt = Date.now()

      // If an unblocked is already logged that matches loop_id + callback_cntr
      // (possibly faked) then this is out of order delivery and needs to be
      // discarded
      const cacheKey = computeCacheKey(evt)

      if (this._actionCache._blocked.has(cacheKey)) {
        // out of order delivery, delete the cache entry
        const value = this._actionCache._blocked.get(cacheKey)
        if (value.type === 'unblocked') {
          // Out of order delivery
          Logger.info(`Out of order unblocked message processed for agent ${evt.agent}, ignoring`)
        } else {
          // ??? multiple blocked for same tick ???
          Logger.info(`Unexpected double blocked message processed for agent ${evt.agent}, ignoring`)
        }
        this._actionCache._blocked.delete(cacheKey)
      } else {
        const cacheEntry = {
          type: 'blocked',
          event: evt
        }

        // Regular profile actions do not works here because those only fire at
        // unblock, and we want this to run while the block is occurring.
        this.startBlockedProfile(evt.agent, cacheEntry)

        this._actionCache._blocked.upsert(cacheKey, cacheEntry)
      }
    })

    // Now we have an accurate count of the time the process was blocked
    this._context.eventManager.on('process-unblocked', evt => {
      const cacheKey = computeCacheKey(evt)

      const blockedEvent = this._actionCache._blocked.get(cacheKey)
      if (!blockedEvent) {
        // Record that we got an unblocked-without-blocked that may be due to
        // out of order delivery.
        this._actionCache._blocked.upsert(cacheKey, { type: 'unblocked' })
        return
      }

      // Don't need this anymore
      this._actionCache._blocked.delete(cacheKey)

      const agent = this.getAgent(evt)
      this.sendProcessBlocked(agent, {
        blockedFor: evt.args.blockedFor,
        stack: blockedEvent.event.args.stack,
        threshold: blockedEvent.event.args.threshold
      })

      this.endBlockedProfile(blockedEvent, evt.args.blockedFor)
    })

    // for process-blocked events that we never receive a process-unblocked
    // event for or whose unblocked events come in too late.
    this._actionCache._blocked.on('expired', (cacheKey, entry) => {
      // Expiring unblocked-first events are a no-op
      if (entry.type === 'unblocked') return

      // Don't send unblock timeouts for dead agents, because we don't know the
      // actual duration and it's dead anyway.  This shouldn't happen if we got
      // the exit event, but in some cases (like a segfault), that might not
      // happen.
      const agent = this._context.agentManager.byId(entry.event.agent)
      if (!agent) return

      this.sendProcessBlocked(agent, {
        stack: entry.event.args.stack,
        threshold: this._actionCache._blocked._ttlMs
      })
    })

    this._context.diskAssetManager.on('add', asset => {
      if (asset.type === 'snapshot') return

      asset = lodash.clone(asset)
      asset.url = this._context.diskAssetManager.getAssetUrl(asset.asset)
      if (asset.type === 'snapshot-summary') {
        asset.type = 'snapshot'
      }

      const agent = this._context.agentManager.byId(asset.id) ||
        this._context.agentManager.deadById(asset.id)
      if (agent && asset) {
        this.sendAssetGenerated(agent, asset)
      }
    })

    this._context.eventManager.on('new-vulnerability-found', evt => {
      const vulnId = evt.args.vulnId
      const agents = this._context.agentManager.byVulnId(evt.args.vulnId)

      // if for some reason no agents found, return -- perhaps it was a very
      // short-lived process?  Gonna log it tho, since it's unexpected.
      if (agents.length === 0) {
        Logger.warn(`no agents found for new vulnerability ${vulnId}, not notifying`)
        return
      }

      this.sendVulnerabilityDetected(agents[0], vulnId)
    })

    setImmediate(() => {
      this.updateActionsList()
      this.updateNotificationsList()
      this.updateIntegrations()
      this.runViews()
    })
  }

  updateActionsList () {
    this._context.settingsManager.getSettings('savedViews', (err, views) => {
      if (err) {
        Logger.warn('Could not update actions list.', err)
        return
      }

      const oldViews = this._savedViews
      this._savedViews = new Map()

      Object.keys(views).forEach(key => {
        let view = views[key]
        view = lodash.clone(view)
        view.expression = qe.fromJSON(this._context, view.expression)
        view.url = `${this._context.config.notification.consoleURL}/processes?view=${view.id}`
        view.consoleURL = this._context.config.notification.consoleURL

        const oldView = oldViews.get(view.id)
        view.running = oldView ? oldView.running : false

        this._savedViews.set(view.id, view)
      })

      // Delete state for views that have been deleted
      for (let viewId of oldViews.keys()) {
        if (!this._savedViews.has(viewId)) {
          this._actionCache.deleteState({ view: { id: viewId } })
        }
      }

      Logger.debug(`Monitoring ${this._savedViews.size} saved views`)
    })
  }

  updateIntegrations () {
    this._context.settingsManager.getSettings('integrations', (err, integrations) => {
      if (err) throw err

      this._savedViews.forEach(view => {
        if (!view.actions) return

        view.actions = view.actions.map(action => {
          if (!action.integrationId) return action
          return integrations[action.integrationId] || action
        })
        this._savedViews.set(view.id, view)
      })

      this._notifications.forEach(notif => {
        if (!notif.actions) return

        notif.actions = notif.actions.map(action => {
          if (!action.integrationId) return action
          return integrations[action.integrationId] || action
        })
        this._notifications.set(notif.event, notif)
      })
    })
  }

  // There are a very small number of possible events limited by the console,
  // so no need to constantly be pruning the cache. They can only ever
  // take up a few kb of memory. - Famous Last Words
  updateNotificationsList () {
    this._context.settingsManager.getSettings('notifications', (err, notifs) => {
      if (err) {
        Logger.warn('Could not update notifications.', err)
        return
      }

      this._notifications.clear()

      Object.keys(notifs).forEach(event => {
        const notif = lodash.clone(notifs[event])
        this._notifications.set(notif.event, notif)
      })
    })
  }

  // Loop through all saved views. If actions are configured, run the query.
  // If agents are returned, run the actions. Ignore action errors to keep from
  // blocking subsequent actions from running.
  runViews () {
    const now = Date.now()
    const agentManager = this._context.agentManager

    const viewTasks = []
    this._savedViews.forEach((view, id) => {
      viewTasks.push(vcb => {
        // Run the query. Don't do anything if no matches.
        const agents = agentManager.query(view.expression)
        const params = { view, agents, now, event: 'nsolid-saved-view-match' }
        if (!this.shouldRunSavedView(params)) return setImmediate(vcb)

        Logger.debug(`${agents.length} agents matching ${view.name}`, view.actions)
        view.running = true

        this.runEventActions(params, err => {
          // Actions may still running at this point so don't set the flag to
          // false.  The async.parallel callback will set that once everything
          // is complete.

          this._actionCache.setLastRun(id, params)
          vcb(err)
        })
      })
    })

    // Run each view serially to space the actions out some.
    // TODO evaluate whether this is desireable
    async.series(viewTasks, err => {
      if (err) Logger.debug('Error running saved views', err)
      setTimeout(() => this.runViews(), this.interval)
    })
  }

  runEventActions (params, cb) {
    // Collect all the asset information from profiling and snapshotting
    // At this time only cpu profiles and heap snapshots are run per-agent.
    const profileTasks = []
    const snapshotTasks = []
    // At the moment, only notification actions need the agent
    // to be formatted.
    params.agents = params.agents.map(agent => {
      agent = formatAgent(agent)

      // Never allow asset created events to trigger the creation of other assets.
      if (!params.artifacts) {
        profileTasks.push(this.buildTask(lodash.defaults({agent}, params), this._actions.profile))
        snapshotTasks.push(this.buildTask(lodash.defaults({agent}, params), this._actions.snapshot))
      }

      return agent
    })

    // We purposefully ignore errors here in order to not block other actions
    async.parallel({
      profiles: acb => async.parallel(profileTasks, acb),
      snapshots: acb => async.parallel(snapshotTasks, acb)
    }, (err, results) => {
      if (err) Logger.warn(`Error running actions for ${params.view.name || params.event}`, err)
      // Run all the notifications tasks
      // results is { profiles: [], snapshots: [] }
      params.artifacts = flattenArtifacts(this._context, params.artifacts || results)
      async.parallel([
        this.buildTask(params, this._actions.email),
        this.buildTask(params, this._actions.slack),
        this.buildTask(params, this._actions.webhook)
      ], err => {
        if (err) Logger.warn(`Error running notifications for ${params.view.name || params.event}`, err)

        // The cloned view object we are using may have since been replaced in
        // the cache, so get the current one and mark it not running.
        const currentView = this._savedViews.get(params.view.id)
        if (currentView) currentView.running = false
      })
    })

    // The task is scheduled and the view is "running", return control flow to
    // the saved view system
    setImmediate(cb)
  }

  buildTask (params, actionRunner) {
    // Don't pass err to the callback. It will block subsequent actions
    return acb => {
      actionRunner.run(params, (err, res) => {
        if (err) {
          Logger.warn(`${actionRunner.type} failed on '${params.view.name || params.event}'`, err)
        }

        acb(null, res)
      })
    }
  }

  shouldRunGlobalNotification (params) {
    // Don't do anything if no actions configured.
    // No snooze or delay for global notifications.
    if (!viewHasActions(params.view)) return false
    return params.agents.length !== 0
  }

  shouldRunSavedView (params) {
    // Don't run a saved view if it's still running from a previous iteration
    if (params.view.running) return false

    // Don't do anything if no actions configured.
    if (!viewHasActions(params.view)) return false

    if (!params.agents.length) {
      this._actionCache.setEmpty(params.view.id, params.view)
      return false
    }
    this._actionCache.setEnter(params.view.id, params)

    // Don't do anything if we're in snooze or delay
    return this._actionCache.shouldRun(params.view.id, params)
  }

  // Not used ATM as this global event was removed from the Console in 3.x
  sendAgentFound (agent) {
    const event = 'nsolid-agent-found'
    this.sendGlobalEvent(agent, event)
  }

  sendProcessBlocked (agent, data) {
    const event = 'nsolid-process-blocked'
    const config = this._notifications.get(event)
    if (!config) return

    // The runtime should discard stack samples that happen to occur during
    // metrics execution.  But if one gets through, don't expose them to users
    // because the optics are pretty bad.
    if (/nsolid\.js:/.test(data.stack)) return

    // `stack` always comes from the process-blocked event, and we'll add
    // a more customized human readable message anyway, so just always
    // remove the first line (the message) from the stack.
    data.stack = data.stack.replace(/.*\n/, '')

    if (config.threshold && (data.blockedFor || data.threshold) < config.threshold) return

    this.sendGlobalEvent(agent, event, null, data)
  }

  startBlockedProfile (agent, cacheEntry) {
    const event = 'nsolid-process-blocked'
    const config = this._notifications.get(event)
    if (!config) return
    if (!config.profileWhileBlocked) return

    const profileMinDuration = config.profileMinDuration || 1000

    const assetCommand = {
      command: 'profile',
      id: cacheEntry.event.agent,
      args: {
        duration: 60000,
        metadata: {
          reason: 'Blocked Event Loop'
        }
      },
      profileMinDuration
    }

    cacheEntry.profile = httpUtils.runAssetCommand(assetCommand, this._context, () => {
      // If profile is complete, remove its requestId
      cacheEntry.profile = null
    })
    cacheEntry.profileMinDuration = profileMinDuration
  }

  endBlockedProfile (cacheEntry, blockedFor) {
    if (cacheEntry.profile) {
      if (blockedFor >= cacheEntry.profileMinDuration) {
        httpUtils.endProfile(this._context, cacheEntry.profile, cacheEntry.event.agent)
      } else {
        httpUtils.cancelProfile(this._context, cacheEntry.profile, cacheEntry.event.agent)
      }
    }
  }

  sendAgentExit (agent, data, profileId) {
    const event = 'nsolid-agent-exit'

    // If a profile was saved at agent exit, include its data
    var assets = null
    if (profileId) {
      // Fake the asset object because it may not actually exist yet
      assets = { profiles: [ { type: 'profile', asset: profileId } ] }
    }

    this.sendGlobalEvent(agent, event, assets, data)
  }

  sendAssetGenerated (agent, asset) {
    const event = `nsolid-${asset.type}-generated`
    const artifacts = {}
    artifacts[`${asset.type}s`] = [asset]
    this.sendGlobalEvent(agent, event, artifacts)
  }

  sendGlobalEvent (agent, event, artifacts, data) {
    // Not really a saved view, but it has some of the same shape
    const view = this._notifications.get(event)
    const now = Date.now()
    const consoleURL = this._context.config.notification.consoleURL
    const params = { consoleURL, event, now, view, agents: [formatAgent(agent)], artifacts, data }
    if (!view || !this.shouldRunGlobalNotification(params)) return

    this.runEventActions(params, err => {
      if (err) Logger.warn(`Error running notifications for ${event}`, err)
    })
  }

  sendVulnerabilityDetected (agent, vulnId) {
    const event = 'nsolid-vulnerability-detected'
    const vuln = Object.assign({}, this._context.vulnManager.getVulnById(vulnId))
    vuln.url = `${this._context.config.notification.consoleURL}/vulnerability/${vulnId}`
    this.sendGlobalEvent(agent, event, null, { vulnerabilities: [vuln] })
  }

  // Build a duck.
  getAgent (evt) {
    const agent = this._context.agentManager.byId(evt.agent)
    if (agent) return agent
    return {id: evt.agent, info: evt.args}
  }
}

function viewHasActions (view) {
  return (view && view.actions && view.actions.length > 0)
}

function flattenArtifacts (context, results) {
  const artifactsTypes = ['profiles', 'snapshots']

  const artifacts = []
  artifactsTypes.forEach(type => {
    if (!results[type]) return

    results[type].forEach(asset => {
      if (!asset) return

      artifacts.push({
        type: asset.type,
        url: context.diskAssetManager.getAssetUrl(asset.summaryAsset || asset.bulkId || asset.asset, asset.type)
      })
    })
  })
  return artifacts
}

function formatAgent (agent) {
  return lodash.pick(agent, ['id', 'info', 'metrics'])
}

function computeCacheKey (evt) {
  const agentId = evt.agent
  const loopId = evt.args.loopId
  const callbackCntr = evt.args.callbackCntr || 0
  const cacheKey = `${agentId}-${loopId}-${callbackCntr}`
  return cacheKey
}
