'use strict'

exports.getConfig = getConfig
exports.getDefaultEnvVars = getDefaultEnvVars

const path = require('path')
const constants = require('constants')

const toml = require('toml')
const minimist = require('minimist')
const lodash = require('lodash')
const userHome = require('user-home')
const stringCase = require('stringcase')

const util = require('./util')

const Logger = util.getLogger(__dirname, __filename)

exports.SampleConfigFile = path.join(__dirname, '..', 'config-sample.toml')

// config property paths which are files resolved relative to config file path
// for config files, cwd for env vars
const RelativePropertyPaths = [
  'web.https.key',
  'web.https.cert',
  'data.dir',
  'logs.influx'
]

const argvOpts = {
  alias: {
    'h': 'help',
    'v': 'version',
    'p': 'print-config',
    'c': 'config',
    'r': 'relay',
    'd': 'provider'
  },
  strings: [
    'config', 'provider'
  ]
}
const Argv = minimist(process.argv.slice(2), argvOpts)

const ConfigDefault = require('./config-default')
resolveRelativePaths(ConfigDefault, process.cwd())

const ConfigExtra = require('./config-extra')

const PathIsJson = new Set([
  'sockets.secondaryRemotes'
])

// Return an object describing the configuration.
function getConfig () {
  if (Argv.help || (Argv._[0] === '?')) return { do_help: true }
  if (Argv.version) return { do_version: true }

  const envConfig = getConfigFromEnv()
  const envStorageConfig = getConfigFromEnv(true)
  const stdConfig = getStdConfigFromEnv()

  let fileConfig

  if (Argv.config) {
    fileConfig = getConfigFromFile(Argv.config)
    if (fileConfig == null) return null
    for (let path of PathIsJson) {
      if (lodash.has(fileConfig, path)) {
        const parsed = JSON.parse(lodash.get(fileConfig, path))
        lodash.set(fileConfig, path, parsed)
      }
    }
  }

  const config = lodash.defaultsDeep(envConfig, stdConfig, fileConfig, envStorageConfig, ConfigDefault, ConfigExtra)
  coerceConfig(config, ConfigDefault)
  coerceConfig(config, ConfigExtra)

  // fix up anything needed in the final config
  delete config.fileName

  // set http proxies to null if currently empty string
  if (config.web.httpProxy === '') config.web.httpProxy = null
  if (config.web.httpsProxy === '') config.web.httpsProxy = null

  // if http proxy set, but not https proxy, set https proxy to http proxy
  if (config.web.httpsProxy == null && config.web.httpProxy != null) {
    config.web.httpsProxy = config.web.httpProxy
  }

  // use key/crt in ../etc if the relevant values are blank
  const etcPath = path.resolve(path.join(__dirname, '..', 'etc'))
  if (config.web.https.key === '') {
    config.web.https.key = path.join(etcPath, 'nsolid-self-signed.key')
  }

  if (config.web.https.cert === '') {
    config.web.https.cert = path.join(etcPath, 'nsolid-self-signed.crt')
  }

  if (config.web.https.secureOptions) {
    parseSecureOptions(config)
  }

  if (config.data.retentionDays <= 0) config.data.retentionDays = 1

  config.isStaging = (process.env.NSOLID_STAGING != null)

  if (Argv['print-config']) {
    config.do_configPrint = true
    return config
  }

  if (Argv['generate-keypair']) {
    config.do_genkey = true
    return config
  }

  if (Argv['repair-db']) {
    config.do_repairDb = true
    return config
  }

  if (Argv.relay) {
    config.relayMode = true
  }

  config.license.provider = (Argv.provider || 'default').toLowerCase()

  if (util.ensureDirExists(config.data.dir)) return null

  return config
}

// Return the env vars supported, with values from the default.  Do not return
// variables defined in config-extra, as this is used to print help
function getDefaultEnvVars () {
  const result = {}

  for (let path of getPaths(ConfigDefault)) {
    if (path === 'fileName') continue

    const envVar = configPathToEnvVar(path)
    const defVal = lodash.get(ConfigDefault, path)

    result[envVar] = defVal
  }

  return result
}

// Get configuration from "standard" env vars.
function getStdConfigFromEnv () {
  const config = { web: {} }

  if (process.env.http_proxy) config.web.httpProxy = process.env.http_proxy
  if (process.env.HTTP_PROXY) config.web.httpProxy = process.env.HTTP_PROXY
  if (process.env.https_proxy) config.web.httpsProxy = process.env.https_proxy
  if (process.env.HTTPS_PROXY) config.web.httpsProxy = process.env.HTTPS_PROXY

  if (process.env.PORT) config.web.server = `:${process.env.PORT}`

  return config
}

// Get the configuration from env vars, from the paths.
function getConfigFromEnv (useStorageNames) {
  const config = {}

  for (let path of getPaths(ConfigDefault)) {
    const envVarName = configPathToEnvVar(path, useStorageNames)
    let envVarValue = process.env[envVarName]
    if (envVarValue == null) continue

    if (PathIsJson.has(path)) {
      envVarValue = JSON.parse(envVarValue)
    }

    lodash.set(config, path, envVarValue)
  }

  for (let path of getPaths(ConfigExtra)) {
    const envVarName = configPathToEnvVar(path, useStorageNames)
    let envVarValue = process.env[envVarName]
    if (envVarValue == null) continue

    if (PathIsJson.has(path)) {
      envVarValue = JSON.parse(envVarValue)
    }

    lodash.set(config, path, envVarValue)
  }

  resolveRelativePaths(config, process.cwd())

  return config
}

function configPathToEnvVar (path, useStorageNames) {
  let constPath = stringCase.constcase(path.replace(/\./g, '_'))
  constPath = constPath
    .replace('_H_W_M', '_HWM')
    .replace('_U_R_L', '_URL')

  return `NSOLID_${useStorageNames ? 'STORAGE' : 'CONSOLE'}_${constPath}`
}

// Get the configuration specified in a file.
function getConfigFromFile (tomlFileName) {
  let configToml = getConfigFileToml(tomlFileName)
  if (configToml == null) return null

  let config
  try {
    config = toml.parse(configToml)
  } catch (err) {
    Logger.error(err, `error parsing toml config: ${tomlFileName}:${err.line}:${err.column}`)
    return null
  }

  resolveRelativePaths(config, path.dirname(tomlFileName))

  config.fileName = tomlFileName

  return config
}

// Get the TOML content in the spcified file.
function getConfigFileToml (fileName) {
  if (!util.fileExists(fileName)) {
    Logger.error(`config file does not exist: ${fileName}`)
    return null
  }

  const configToml = util.loadFile(fileName)
  if (configToml == null) {
    Logger.error(`error reading config file: "${fileName}"`)
    return null
  }

  return configToml
}

// Get the "paths" of the default config, for use w/lodash.
function getPaths (config, prefix, result) {
  if (result == null) result = []
  if (prefix == null) prefix = ''

  for (let key in config) {
    const val = config[key]
    if (lodash.isObject(val) && !lodash.isArray(val)) {
      getPaths(val, `${prefix}${key}.`, result)
    } else {
      result.push(`${prefix}${key}`)
    }
  }

  return result
}

// Resolve relative path properties against a base directory
function resolveRelativePaths (config, baseDir) {
  for (let propertyPath of RelativePropertyPaths) {
    let relPath = lodash.get(config, propertyPath)
    if (relPath == null) continue

    relPath = `${relPath}`
    if (relPath === '') continue

    relPath = relPath.replace(/^~/, userHome)

    lodash.set(config, propertyPath, path.resolve(baseDir, `${relPath}`))
  }
}

// Coerce the values in config to be the same type as values in ConfigDefault.
function coerceConfig (config, baseConfig) {
  const propPaths = getPaths(baseConfig)

  for (let propPath of propPaths) {
    const type = typeof lodash.get(baseConfig, propPath)

    // get the value from the config, if it's the same type as default,
    // no coercion required
    let value = lodash.get(config, propPath)
    if (typeof value === type) continue // eslint-disable-line valid-typeof

    let origValue = value

    // strings coerce very easily
    if (type === 'string') {
      value = `${value}`
    // numbers should be coerced to integers
    } else if (type === 'number') {
      value = parseInt(`${value}`.trim())
      if (isNaN(value)) {
        Logger.error(`config value for ${propPath}, '${origValue}', must be a number; using default`)
        value = lodash.get(baseConfig, propPath)
      }
    // Unexpected type!
    // Shouldn't happen in practice; probably only boolean, only with hacked
    // default configuration.  So complain.
    } else {
      Logger.error(`config value for ${propPath}, '${origValue}', has unexpected type '${type}'; using default`)
      value = lodash.get(baseConfig, propPath)
    }

    // if we made it this far, value was coerced, so update in the config
    lodash.set(config, propPath, value)
  }

  return config
}

function parseSecureOptions (config) {
  let secureOptionsMask = 0
  const secureOptions = config.web.https.secureOptions.split('|')
  for (let secureOption of secureOptions) {
    secureOption = secureOption.trim()
    const secureOptionConstant = constants[secureOption]
    if (!secureOptionConstant) {
      Logger.error(`Unknown secureOptions value specified: ${secureOption}`)
      process.exit(1)
    }
    secureOptionsMask = secureOptionsMask | secureOptionConstant
  }
  config.web.https.secureOptionsMask = secureOptionsMask
}
