'use strict'

const path = require('path')
const EventEmitter = require('events')
const lodash = require('lodash')
const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const async = require('async')
const levelup = require('levelup')
const sub = require('level-sublevel')
const leveldown = require('leveldown')

exports.create = create
exports.repairDb = repairDb

const types = ['savedViews', 'assets', 'integrations', 'license', 'agents', 'generalSettings', 'notifications']
const noIdTypes = ['license']
const noOpFunction = function noop () {}
const template = {
  integrations: {},
  generalSettings: {},
  notifications: {},
  agents: {},
  savedViews: {},
  assets: {},
  license: {}
}
function create (context) {
  return new LevelDBConnector(context)
}

function dbFile (config) {
  return path.join(config.data.dir, config.data.dbname)
}

// Manages the application settings.
class LevelDBConnector extends EventEmitter {
  constructor (context) {
    super()

    this._context = context
    this._file = dbFile(context.config)
    const leveldb = levelup(this._file)
    leveldb.on('error', err => {
      setImmediate(() => {
        throw new Error(`Error accessing settings database, is another instance of nsolid-console running?  LevelDB error: ${err.type}`)
      })
    })
    leveldb.on('open', _ => {
      this.emit('connect')
    })
    this._db = sub(leveldb)
    this._settingsDb = this._db.sublevel('settings')
    this._assetsDb = this._db.sublevel('assets')
    this._agentsDb = this._db.sublevel('agents')
    this._settings = lodash.clone(template)

    this.setupDbApi()
  }

  // Return a subset of the current settings.
  getSettings (type, cb) {
    setImmediate(cb, null, lodash.cloneDeep(this._settings[type]))
  }

  // Returns the agents stored in leveldb, in the callback.
  // The callback is invoked for every agent returned from leveldb.
  // On error, the callback is invoked with the error from leveldb, and
  // no more agents will be returned.
  // When there are no more agents to be returned, a null will be passed
  // to the callback instead of an agent.
  streamAgents (cb) {
    this._agentsDb.createReadStream()
      .on('data', data => {
        if (cb != null) cb(null, data)
      })
      .on('error', err => {
        Logger.error(err, 'error reading agents from leveldb')
        if (cb != null) cb(err)
        cb = null
      })
      .on('end', () => {
        if (cb != null) cb()
        cb = null
      })
      .on('close', () => {
        if (cb != null) cb()
        cb = null
      })
  }

  // Get a single entry by ID
  getEntry (type, id, cb) {
    if (!type) {
      return setImmediate(cb, new Error('Missing required argument `type`.'))
    } else if (noIdTypes.indexOf(type) === -1 && !id) {
      return setImmediate(cb, new Error('Missing required argument `id`.'))
    } else if (types.indexOf(type) === -1 && noIdTypes.indexOf(type) === -1) {
      return setImmediate(cb, new Error('Invalid value for required argument `type`.'))
    }

    const res = this.get[type](this._settings, type, id)
    setImmediate(cb, null, res)
  }

  setupDbApi () {
    this.delete = {}
    this.delete.integrations = buildDeleter(deleteObjType, 'id', this._settingsDb).bind(this)
    this.delete.generalSettings = buildDeleter(deleteObjType, 'id', this._settingsDb).bind(this)
    this.delete.notifications = buildDeleter(deleteObjType, 'event', this._settingsDb).bind(this)
    this.delete.agents = buildDeleter(noOpFunction, 'id', this._agentsDb).bind(this)
    this.delete.savedViews = buildDeleter(deleteObjType, 'id', this._settingsDb).bind(this)
    this.delete.assets = buildDeleter(deleteObjType, 'id', this._assetsDb).bind(this)
    this.delete.license = buildDeleter(deleteNoIdType, null, this._settingsDb).bind(this)

    this.set = {}
    this.set.integrations = buildSetter(setObjType, 'id', this._settingsDb).bind(this)
    this.set.generalSettings = buildSetter(setObjType, 'id', this._settingsDb).bind(this)
    this.set.notifications = buildSetter(setObjType, 'event', this._settingsDb).bind(this)
    this.set.agents = buildSetter(noOpFunction, 'id', this._agentsDb).bind(this)
    this.set.savedViews = buildSetter(setObjType, 'id', this._settingsDb).bind(this)
    this.set.assets = buildSetter(setObjType, 'id', this._assetsDb).bind(this)
    this.set.license = buildSetter(setNoIdType, null, this._settingsDb).bind(this)

    this.get = {}
    this.get.integrations = getObjType
    this.get.generalSettings = getObjType
    this.get.notifications = getObjType
    this.get.license = getNoIdType
    this.get.agents = noOpFunction
    this.get.savedViews = getObjType
    this.get.assets = getObjType

    // This are only used for loading settings initially
    this.load = {}
    this.load.integrations = setObjType
    this.load.generalSettings = setObjType
    this.load.notifications = setObjType
    this.load.license = setNoIdType
    this.load.agents = noOpFunction
    this.load.assets = setObjType
    this.load.savedViews = setObjType
  }

  deleteSetting (type, id, cb) {
    if (!type) {
      return setImmediate(cb, new Error('Missing required argument `type`.'))
    } else if (noIdTypes.indexOf(type) === -1 && !id) {
      return setImmediate(cb, new Error('Missing required argument `id`.'))
    } else if (types.indexOf(type) === -1 && noIdTypes.indexOf(type) === -1) {
      return setImmediate(cb, new Error('Invalid value for required argument `type`.'))
    }

    this.delete[type](type, id, (err, settings) => {
      if (err) return cb(err)
      this._settings = settings
      cb()
    })
  }

  readSettings (cb) {
    const settings = lodash.clone(template)
    async.parallel([
      acb => this.readSublevel(this._settingsDb, settings, acb),
      acb => this.readSublevel(this._assetsDb, settings, acb)
    ], err => {
      if (err) return cb(err)
      if (settings) this._settings = settings

      cb(null, settings)
    })
  }

  setSetting (type, item, cb) {
    if (!type) {
      return setImmediate(cb, new Error('Missing required argument `type`.'))
    } else if (!item) {
      return setImmediate(cb, new Error('Missing required argument `item`.'))
    } else if (types.indexOf(type) === -1 && noIdTypes.indexOf(type) === -1) {
      return setImmediate(cb, new Error('Invalid value for required argument `type`.'))
    }

    this.set[type](type, item, (err, settings) => {
      if (err) return cb(err)
      this._settings = settings

      this._context.settingsManager.emitSettingsEvent(type, {type, item})
      cb(null, item)
    })
  }

  readSublevel (db, settings, cb) {
    cb = lodash.once(cb)
    db.createReadStream()
      .on('data', data => {
        this.parseSetting(data, settings)
      })
      .on('error', err => {
        replyError(err, err => {
          Logger.error(err)
          cb(err)
        })
      })
      .on('end', () => {
        this._settings = settings
        cb()
      })
      .on('close', () => {
        Logger.debug('Done loading settings.', this._settings)
      })
  }

  /*
   * Mutate the settings object, adding read settings data to it.
   */
  parseSetting (data, settings) {
    var value
    try {
      value = JSON.parse(data.value.toString())
    } catch (err) {
      Logger.warn('Settings parse error.', err)
      return
    }

    const parts = data.key.split(':')
    const type = parts[0]
    const id = parts[1]

    if (!this.load[type]) return

    try {
      this.load[type](settings, type, value, id)
    } catch (err) {
      Logger.warn('LevelDB parsing error', data, err)
    }
  }

  /*
   * Different types of data use a different format for their key.
   *
   * Currently we only have one setting with a key that is a static string.
   * There can only be one license, so there is no use having `license:${key}`
   */
  buildKey (type, item) {
    if (noIdTypes.indexOf(type) > -1) {
      return type
    }

    var idAttr
    switch (type) {
      case 'notifications':
        idAttr = 'event'
        break
      case 'integrations':
      case 'savedViews':
      case 'assets':
      case 'generalSettings':
      default:
        idAttr = 'id'
        break
    }

    if (!idAttr || !item[idAttr]) {
      throw new Error('Invalid settings format.')
    }

    return `${type}:${item[idAttr]}`
  }
}

/*
 * Convert LevelDB's incredibly stupid custom errors to standard
 * Node callback arguments & errors and call the passed callback as such.
 */
function replyError (err, cb) {
  switch (err.type) {
    case 'NotFoundError':
      return cb()
    case 'WriteError':
    case 'ReadError':
    case 'EncodingError':
    case 'OpenError':
    case 'InitializationError':
    case 'LevelUPError': // Error base class
    default:
      return cb(new Error(`DB Error: ${err.type}. ${JSON.stringify(err)}`))
  }
}

/*
 * Read from cache
 */
function getObjType (settings, type, id) {
  return settings[type][id]
}

function getNoIdType (settings, type) {
  return settings[type]
}

/*
 * Build functions to handle saving different types of data
 * to the database.
 */
function setObjType (settings, type, item, id) {
  settings[type][id] = item
}

function setNoIdType (settings, type, item) {
  settings[type] = item
}

/*
 * Functions to handle deleting different types of data
 * from the database.
 */
function deleteObjType (settings, type, id) {
  delete settings[type][id]
}

function deleteNoIdType (settings, type) {
  delete settings[type]
}

// attempt to repair the leveldb bits, call cb with error on failure
function repairDb (config, cb) {
  const fileName = dbFile(config)

  console.log(`attempting leveldown repair of ${fileName}`)

  try {
    leveldown.repair(fileName, repairDone)
  } catch (err) {
    console.log(`unsuccessful leveldown repair of ${fileName}: ${err}`)
    return setImmediate(cb, err)
  }

  function repairDone (err) {
    if (err == null) {
      console.log(`successful leveldown repair of ${fileName}`)
      return cb()
    }

    console.log(`unsuccessful leveldown repair of ${fileName}: ${err}`)
    cb(err)
  }
}

function buildDeleter (deleteFromCache, idAttr, db) {
  return function deleter (type, id, cb) {
    var key
    const item = {}

    if (idAttr) item[idAttr] = id

    try {
      key = this.buildKey(type, item)
    } catch (err) {
      return setImmediate(cb, err)
    }

    db.del(key, err => {
      if (err) return replyError(err, cb)

      deleteFromCache(this._settings, type, id)

      cb(null, this._settings)
    })
  }
}

function buildSetter (setOnCache, idAttr, db) {
  // Will get bound to the LevelDBConnector
  return function setter (type, item, cb) {
    var id
    var key
    if (idAttr) id = item[idAttr]
    try {
      key = this.buildKey(type, item)
    } catch (err) {
      return setImmediate(cb, err)
    }

    db.put(key, JSON.stringify(item), err => {
      if (err) return replyError(err, cb)

      setOnCache(this._settings, type, item, id)

      cb(null, this._settings)
    })
  }
}
