'use strict'

const lodash = require('lodash')

const httpUtils = require('./http-utils')

exports.run = run

/*
 * Run a command and write each response to the event stream. If an error
 * occurs, end the stream. If we get an `end` event from the command dispatcher,
 * end the stream.
 *
 * Instead of using an event stream, a caller can send an httpResponse object,
 * from the server, to write the first response returned from the command as
 * the only output.
 *
 * @param {object} context
 * @param {EventStream} eventStream
 * @param {object} params
 * @param {function} params.transformer - A function that gets passed a ZeroMQ
 * message, synchronously or asynchronously transforms the message, and passes
 * it to a node-style callback.
 * @returns {string} requestId
 */
const noopTransformer = (data, cb) => cb(null, data)

function run (context, eventStream, httpResponse, qData, params) {
  const command = params.command || 'unknown command'

  if (!eventStream && !httpResponse) {
    throw new Error('Can not run a simple command without event stream or httpResponse.')
  }

  if (httpResponse) eventStream = new OneShotEventStream(httpResponse)

  const transformer = params.transformer || noopTransformer
  params = lodash.clone(params)
  delete params.transformer

  if (qData.id) {
    const agent = context.agentManager.byId(qData.id)
    if (!agent) {
      eventStream.writeError(null, `command ${command}`, `No process with the id ${qData.id}.`)
      return endCommand(null, eventStream, context)
    }
    params.id = agent.id
  } else {
    const filter = httpUtils.getBroadcastFilter(qData)
    if (filter != null) {
      params.filter = filter
    }
  }

  const requestId = context.commandDispatcher.send(params)

  context.commandDispatcher.on(`error:${requestId}`, err => {
    eventStream.writeError(null, `command ${command}`, err.message)
    return endCommand(requestId, eventStream, context)
  })

  context.commandDispatcher.on(`data:${requestId}`, data => {
    if (eventStream.isClosed()) {
      return endCommand(requestId, eventStream, context)
    }

    if (data) {
      if (data.body === 'PONG') data.body = {pong: true}
      const responseBody = data.body ? lodash.clone(data.body) : {}
      responseBody.id = data.agentId
      responseBody.time = new Date(data.time)

      transformer(responseBody, (err, transformedData) => {
        if (err) {
          eventStream.writeError(null, `command ${command}`, err.message)
          return endCommand(requestId, eventStream, context)
        }

        eventStream.writeData(transformedData)
        // All POST commands at this point only expect one response if an id is supplied
        if (qData.id != null && !eventStream.isClosed()) eventStream.close()
      })
    }
  })

  context.commandDispatcher.on(`end:${requestId}`, () => {
    endCommand(requestId, eventStream, context)
  })

  return requestId
}

function endCommand (id, eventStream, context) {
  eventStream.close()
  context.commandDispatcher.removeAllListeners(`end:${id}`)
  context.commandDispatcher.removeAllListeners(`data:${id}`)
  context.commandDispatcher.removeAllListeners(`error:${id}`)
}

class OneShotEventStream {
  constructor (httpResponse) {
    this.res = httpResponse
  }

  close () {
    if (this.isClosed()) return

    this.res.send({error: 'command terminated'})
    this.res = null
  }

  isClosed () {
    return this.res == null
  }

  writeData (data) {
    this.res.send(data)
    this.res = null
  }

  writeError (ignore, title, message) {
    this.res.send({error: `${title} - ${message}`})
    this.res = null
  }
}
