'use strict'

exports.handleRequest = handleRequest
exports.verifySchema = verifySchema

const fs = require('fs')
const path = require('path')

const markdown = require('markdown').markdown

const API = require('./API')
const pkg = require('../../../package.json')

const VERSION = API.VERSION

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

let Html

function handleRequest (req, res) {
  const context = req.nsConsole.context
  const schema = context.schemas.http[VERSION]
  // const schema = util.readSchema(`http-v${VERSION}`) // read dynamically

  Logger.debug('handling request')

  if (Html == null) Html = buildHtml(context, schema)

  res.set('Content-Type', 'text/html')
  res.send(Html)
}

function buildHtml (context, schema) {
  const htmlTemplateFileName = path.join(__dirname, 'api-docs-template.html')
  const htmlTemplate = fs.readFileSync(htmlTemplateFileName, 'utf8')

  verifySchema(schema)

  const toc = getToc(schema)
  const apis = getApis(schema)
  const parms = getParms(schema)
  const fieldsMetrics = getFields(context, 'metrics')
  const fieldsInfo = getFields(context, 'info')

  return htmlTemplate
    .replace(/%%VERSION%%/g, VERSION)
    .replace(/%%SERVER-VERSION%%/g, pkg.version)
    .replace(/%%TOC%%/g, toc)
    .replace(/%%APIS%%/g, apis)
    .replace(/%%PARMS%%/g, parms)
    .replace(/%%METRICS-FIELDS%%/g, fieldsMetrics)
    .replace(/%%INFO-FIELDS%%/g, fieldsInfo)
}

const NO_DESCRIPTION = 'NO DESCRIPTION!!!'
const NO_DEFINITION = 'NO DEFINITION!!!'
const NO_DEFINITION_OBJECT = {
  desc: 'NO DEFINITION!!!'
}

// create the parms definitions as HTML
function getParms (schema) {
  const allParms = []
  for (let parmName in schema.parm) {
    const parm = schema.parm[parmName]
    parm.name = parmName
    parm.id = `parm-${parmName}`
    allParms.push(parm)
  }

  allParms.sort((a, b) => a.name.localeCompare(b.name))

  const html = []

  for (let parm of allParms) {
    let desc = parm.desc || NO_DESCRIPTION
    if (parm.multiple === true) desc += '<i> (multiple allowed)</i>'
    html.push('<div class="container">')
    html.push(`<h3 class="parm" id="${parm.id}"><a href="#${parm.id}"><tt>${parm.name}</tt></a></h3>`)
    html.push(`<p>${desc}</p>`)
    html.push('')
    if (parm.descLong) {
      html.push(md2html(parm.descLong))
      html.push('')
    }
    html.push('</div>')
  }

  html.push('')

  return html.join('\n')
}

// create the API definitions as HTML
function getToc (schema) {
  const html = []

  // APIs
  const allUris = new Set()
  for (let method in schema) {
    if (method === 'type') continue
    if (method === 'parm') continue

    const apis = schema[method]
    for (let uri in apis) allUris.add(uri)
  }

  html.push('        <h2>APIs</h2>')
  html.push('        <ul>')
  for (let uri of Array.from(allUris).sort()) {
    html.push(`        <li><a href="#uri-${uri}">${uri}</a></li>`)
  }
  html.push('        </ul>')

  // query parms
  html.push('<h2>Query Parms</h2>')
  html.push('        <ul>')

  const allParms = []
  for (let parmName in schema.parm) {
    const parm = schema.parm[parmName]
    parm.name = parmName
    parm.id = `parm-${parmName}`
    allParms.push(parm)
  }

  allParms.sort((a, b) => a.name.localeCompare(b.name))

  for (let parm of allParms) {
    html.push(`        <li><a href="#${parm.id}">${parm.name}</a></li>`)
  }

  html.push('        </ul>')
  return html.join('\n')
}

// create the API definitions as HTML
function getApis (schema) {
  const allApis = []
  for (let method in schema) {
    if (method === 'type') continue
    if (method === 'parm') continue

    const apis = schema[method]
    for (let uri in apis) {
      const api = apis[uri]

      api.uri = uri
      api.method = method
      api.sortKey = `${uri}-${method}`
      api.id = `${method}-${uri}`
      allApis.push(api)
    }
  }

  allApis.sort((a, b) => a.sortKey.localeCompare(b.sortKey))

  const html = []
  let lastUri
  for (let api of allApis) {
    const uriLinkTarget = (lastUri === api.uri) ? '' : ` id="uri-${api.uri}"`
    lastUri = api.uri

    html.push('')
    html.push(`<div class="container"${uriLinkTarget}>`)
    html.push(`<h3 class="api" id="${api.id}"><a href="#${api.id}"><tt>${api.method} /api/v${VERSION}/${api.uri}</tt></a></h3>`)
    html.push('')

    const desc = api.desc || NO_DESCRIPTION
    html.push(`<p>${desc}</p>`)
    html.push('')

    if (api.urlParm) {
      html.push(`<p>The path can also contain the query parameter value for <tt>${api.urlParm}</tt>,`)
      html.push(`eg <tt>${api.method} /api/v${VERSION}/${api.uri}/:${api.urlParm}</tt></p>`)
      html.push('')
    }

    html.push(`<h4>query parameters</h4>`)
    html.push('')
    if (api.parms == null) api.parms = []

    let requiredParm = null

    if (api.parms.length === 0) {
      html.push('<p>none</p>')
    } else {
      html.push('<table width="100%">')
      for (let parm of api.parms) {
        let required = ''
        if (api.required.indexOf(parm) !== -1) {
          required = ' (<b>required</b>)'
          requiredParm = parm
        }
        const parmSchema = schema.parm[parm] || {
          desc: NO_DEFINITION
        }
        let parmDesc = parmSchema.desc || NO_DESCRIPTION
        if (parmSchema.multiple === true) parmDesc += '<i> (multiple allowed)</i>'
        const parmId = `parm-${parm}`
        html.push(`<tr><td class=width-25><tt><a href="#${parmId}"><b>${parm}</b></a></tt> ${required}</td><td>${parmDesc}</td></tr>`)
      }
      html.push('</table>')
    }
    html.push('')

    if (api.request != null) {
      html.push(`<h4>request body</h4>`)
      html.push('')
      const schemaType = schema.type[api.request] || {
        desc: NO_DEFINITION
      }
      const desc = md2html(schemaType.desc) || NO_DESCRIPTION
      html.push(desc)
    }

    let stream = false
    if (api.response != null) {
      html.push(`<h4>response body</h4>`)
      html.push('')

      let responseType = api.response
      if (responseType.startsWith('stream:')) {
        stream = true
        responseType = responseType.substr(7)
      }

      const schemaType = schema.type[responseType] || {
        desc: NO_DEFINITION
      }
      const desc = md2html(schemaType.desc) || NO_DESCRIPTION

      if (stream) {
        let continuous = ''
        if (api.keepalive) {
          continuous = 'The stream is continuous, and will emit output until the client closes the http socket.'
        }
        html.push(`<p>This API returns a stream of JSON objects, described below.  ${continuous}</p>`)
      }
      html.push(desc)
    }

    const rParm = requiredParm ? `'${requiredParm}'` : 'null'
    if (api.tryIt !== false) {
      html.push(`<button onclick="callApi('${api.method}', '${api.uri}', ${stream}, ${rParm})">Try it now!</button>`)
      html.push(`<button onclick="clearOutput('${api.method}', '${api.uri}')">Clear output</button>`)
      html.push(`<pre class="api-output" id="output-${api.method}-${api.uri}"></pre>`)
      html.push('')
    }

    html.push('</div>')
    html.push('')
  }

  return html.join('\n')
}

// create an HTML table list fields for a domain
function getFields (context, domain) {
  const fieldObject = context.fieldManager.fields
  const fields = Object.keys(fieldObject)
    .filter(name => fieldObject[name].domain === domain)
    .sort()
    .map(name => fieldObject[name])

  const html = []
  html.push('<table width="100%">')
  for (let field of fields) {
    let type = field.type
    if (type === 'array') type = 'array of strings'
    html.push(`<tr><td class=width-25><tt>${field.name}</a></tt></td><td>${field.desc} <i>(${type})</i></td></tr>`)
  }
  html.push('</table>')
  return html.join('\n')
}

// verify a schema
function verifySchema (schema) {
  let valid = true
  const types = {}
  const parms = {}

  for (let name in schema.type) {
    const type = schema.type[name]
    if (type.desc == null) {
      Logger.warn(`type ${name} has no "desc" property`)
      type.desc = NO_DESCRIPTION
      valid = false
    }
    types[name] = type
  }

  for (let name in schema.parm) {
    const parm = schema.parm[name]
    if (parm.desc == null) {
      Logger.warn(`parm ${name} has no "desc" property`)
      parm.desc = NO_DESCRIPTION
      valid = false
    }
    parms[name] = parm
  }

  types[NO_DEFINITION] = NO_DEFINITION_OBJECT

  for (let method in schema) {
    if (method === 'type') continue
    if (method === 'parm') continue

    for (let uri in schema[method]) {
      const api = schema[method][uri]
      const messagePrefix = `api ${method} ${uri}`

      if (api.parms == null) api.parms = []
      if (api.required == null) api.required = []

      if (api.response == null) {
        Logger.warn(`${messagePrefix} has no "response" property`)
        api.response = NO_DEFINITION
        valid = false
      }

      const apiParms = {}
      for (let parm of api.parms) {
        if (parms[parm] == null) {
          Logger.warn(`${messagePrefix} parm "${parm}" is not defined`)
          valid = false
        }
        apiParms[parm] = true
      }

      for (let parm of api.required) {
        if (apiParms[parm] == null) {
          Logger.warn(`${messagePrefix} required parm "${parm}" is not defined`)
          valid = false
        }
      }

      if (api.urlParm != null && parms[api.urlParm] == null) {
        Logger.warn(`${messagePrefix} urlParm "${api.urlParm}" is not defined`)
        valid = false
      }

      if (api.urlParm != null && !apiParms[api.urlParm]) {
        Logger.warn(`${messagePrefix} urlParm "${api.urlParm}" is not listed in "parms"`)
        valid = false
      }

      if (api.request && types[api.request] == null) {
        Logger.warn(`${messagePrefix} request type "${api.request}" is not defined`)
        valid = false
      }

      let responseType = `${api.response}`
      if (responseType.startsWith('stream:')) responseType = responseType.substr(7)
      if (types[responseType] == null) {
        Logger.warn(`${messagePrefix} response type "${responseType}" is not defined`)
        valid = false
      }
    }
  }

  return valid
}

function md2html (md) {
  return markdown.toHTML(md, 'Maruku')
}
