'use strict'

exports.handleRequest = handleRequest

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

const asynclib = require('async')

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsConsole.context
  const eventStream = req.nsConsole.eventStream
  const qData = req.nsConsole.qData

  Logger.debug('handling request')

  // set up the date start/end values
  if (qData.start == null) qData.start = Date.now() - 60 * 60 * 1000
  if (qData.end == null) qData.end = Date.now()

  qData.start = util.dateMS2ISO(qData.start)
  qData.end = util.dateMS2ISO(qData.end)

  if (qData.start > qData.end) {
    eventStream.writeError(null, undefined, `start date ${qData.start} greater than end date ${qData.start}`)
    eventStream.close()
    return
  }

  if (context.influxdbReader == null) {
    eventStream.writeError(null, undefined, 'influx is not available')
    eventStream.close()
    return
  }

  const query = buildQuery(context, qData)
  Logger.debug(`query: ${query}`)
  context.influxdbReader.readEvents(query, readEventsDone)

  // handle result of influx query
  function readEventsDone (err, data) {
    if (err) {
      Logger.error(err, 'error reading events')
      eventStream.writeError(null, undefined, 'error reading events')
      eventStream.close()
      return
    }

    if (eventStream.isClosed()) return

    const records = recordsFromQuery(data)
    asynclib.eachSeries(records, (event, eachCb) => {
      setImmediate(() => {
        const record = {
          time: new Date(event.time),
          event: event.type,
          agent: event.agent
        }

        try {
          record.args = JSON.parse(event.body)
        } catch (err) {
          record.args = { error: `error parsing event body: ${err}` }
          Logger.error(err, `error parsing event body: '${event.body}'`)
        }

        eventStream.writeData(record)
        eachCb()
      })
    }, (err) => {
      if (err) {
        Logger.warn(`Error fetching historic events: ${err.message}`)
      }
      eventStream.close()
    })
  }
}

// create metrics records from the query results.
function recordsFromQuery (data) {
  try {
    data = JSON.parse(data)
  } catch (err) {
    Logger.warn(err, 'error parsing JSON data from query', data)
    return []
  }

  Logger.debug(`query results: ${JSON.stringify(data, null, 4)}`)

  if (data.results == null) return []
  if (data.results.length === 0) return []
  if (data.results[0].series == null) return []
  if (data.results[0].series.length === 0) return []

  const series = data.results[0].series[0]
  const records = []

  Logger.debug(`recordsFromQuery(): received ${series.values.length} values`)
  for (let i = 0; i < series.values.length; i++) {
    var record = {}
    for (let j = 0; j < series.columns.length; j++) {
      if (series.values[i][j] == null) continue
      record[series.columns[j]] = series.values[i][j]
    }

    records.push(record)
  }

  return records
}

// build the influx query
function buildQuery (context, qData) {
  const queryParts = ['SELECT']

  queryParts.push('*')
  queryParts.push('FROM rp_events.events WHERE')

  const whereParts = []

  if (qData.type != null) {
    whereParts.push(`type = '${qData.type}'`)
  }

  if (qData.id != null) {
    // normalize id.  to use a short form, the agent must still be running and in cache
    let id, op
    const agent = context.agentManager.byId(qData.id)

    if (agent) {
      id = `'${agent.id}'`
      op = '='
    } else {
      id = `/^${qData.id}/`
      op = '=~'
    }
    whereParts.push(`agent ${op} ${id}`)
  }

  whereParts.push(`time >= '${qData.start}'`)
  whereParts.push(`time <= '${qData.end}'`)

  queryParts.push(whereParts.join(' AND '))

  return queryParts.join(' ')
}
