'use strict'

exports.handleRequest = handleRequest

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

const unlicensedOk = new Set([
  'license-updated'
])

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsConsole.context
  const eventManager = context.eventManager
  const eventStream = req.nsConsole.eventStream

  Logger.debug('handling request')

  const licensed = context.provider && context.provider.license &&
    context.provider.license.valid
  const maxProcesses = context.provider && context.provider.license &&
    context.provider.license.maxProcesses

  // listen for system events
  eventManager.on('system-event', systemEventHandler)

  // Inform this stream whether licensed or not at the beginning of the stream.
  // Create and send an event, but don't emit it, as that will be re-broadcast
  // to everyone, and also written to influxdb.  This should resemble the same
  // event that is sent when the license changes
  const provider = context.provider.type
  const refreshUrl = context.provider.refreshUrl()
  const errorState = context.provider.errorState()
  const orgName = context.provider.orgName()
  const orgAdmins = context.provider.orgAdmins()
  const userRole = req.userRole
  const payload = { licensed: !!licensed, errorState, provider, refreshUrl, userRole, orgName, orgAdmins, maxProcesses }
  const licenseEvent = context.eventManager.createEvent('license-updated', undefined, payload)
  eventStream.writeData(licenseEvent)

  // handle a system event
  function systemEventHandler (event) {
    // client may have closed connection; if so, remove event listener
    if (eventStream.isClosed()) {
      eventManager.removeListener('system-event', systemEventHandler)
      return
    }

    // Freshly get licensed at time of event emit so we can resume publishing
    const licensed = context.provider && context.provider.license &&
      context.provider.license.valid

    if (!licensed && !unlicensedOk.has(event.event)) {
      Logger.debug(`Not sending event ${event.event}`)
      return
    }

    if (event.event === 'license-updated') {
      const orgName = context.provider.orgName()
      const orgAdmins = context.provider.orgAdmins()
      let userRole
      try {
        [ userRole ] = context.provider.userPermitted(req.session)
      } catch (err) {
        // no error if this fails
      }
      event.args = Object.assign({}, event.args, { userRole, orgName, orgAdmins })
    }

    // write the data
    event = Object.assign({}, event)
    event.time = new Date(event.time)
    eventStream.writeData(event)
  }
}
