'use strict'

exports.handleRequest = handleRequest

// seconds
const DEFAULT_INTERVAL = 1

const httpUtils = require('../http-utils')
const normalizers = require('./normalizers')

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsConsole.context
  const eventStream = req.nsConsole.eventStream
  const qData = req.nsConsole.qData
  const interval = qData.interval || DEFAULT_INTERVAL
  const queryExpr = qData.q
  const fields = qData.field.filter(field => field.domain === 'metrics')
  const trackedAgents = new Map()
  let time = new Date()

  Logger.debug('handling request')

  eventStream.writeData({ time: time, type: 'start' })

  poll()
  httpUtils.syncPollTillDone(eventStream, interval * 1000, null, poll)

  function poll () {
    time = new Date()

    eventStream.writeData({ time: time, type: 'interval-start' })

    const nowAgents = context.agentManager.query(queryExpr)
    const newAgents = getNewAgents(trackedAgents, nowAgents)
    const oldAgents = getOldAgents(trackedAgents, nowAgents)

    processOldAgents(time, eventStream, trackedAgents, oldAgents)
    processNewAgents(time, eventStream, trackedAgents, newAgents, fields)
    processExistingAgents(time, eventStream, trackedAgents, newAgents, fields)

    eventStream.writeData({
      time,
      type: 'summary',
      matchingAgents: nowAgents.matchingAgents,
      totalAgents: context.agentManager.count()
    })
    eventStream.writeData({ time: time, type: 'interval-end' })
  }
}

function processOldAgents (time, eventStream, trackedAgents, oldAgents) {
  for (let id of oldAgents.keys()) {
    const record = {
      time: time,
      type: 'agent-leave',
      id: id
    }
    eventStream.writeData(record)
    trackedAgents.delete(id)
  }
}

function processNewAgents (time, eventStream, trackedAgents, newAgents, fields) {
  for (let id of newAgents.keys()) {
    const agent = newAgents.get(id)
    const record = {
      time: time,
      type: 'agent-enter',
      id: id,
      info: normalizers.plain(agent.info),
      metrics: getAgentMetrics(agent, fields)
    }
    eventStream.writeData(record)
    trackedAgents.set(id, agent)
  }
}

function processExistingAgents (time, eventStream, trackedAgents, newAgents, fields) {
  for (let id of trackedAgents.keys()) {
    if (newAgents.has(id)) continue

    const agent = trackedAgents.get(id)

    const record = {
      time: time,
      type: 'metrics',
      id: id,
      metrics: getAgentMetrics(agent, fields)
    }
    eventStream.writeData(record)
  }
}

function getAgentMetrics (agent, fields) {
  if (fields == null || fields.length === 0) {
    return normalizers.plain(agent.metrics || {})
  }

  const result = {}

  for (let field of fields) {
    result[field.name] = field.valueFromAgent(agent)
  }

  return result
}

function getNewAgents (trackedAgents, nowAgents) {
  const newAgents = new Map()

  nowAgents.forEach(agent => {
    if (!trackedAgents.has(agent.id)) newAgents.set(agent.id, agent)
  })

  return newAgents
}

function getOldAgents (trackedAgents, nowAgents) {
  const oldAgents = new Map()

  for (let id of trackedAgents.keys()) {
    const agent = nowAgents.find(agent => agent.id === id)
    if (!agent) oldAgents.set(id, trackedAgents.get(id))
  }

  return oldAgents
}
