'use strict'

exports.handleRequest = handleRequest

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

const LICENSE_KEY_LENGTH = 36

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsConsole.context
  const qData = req.nsConsole.qData
  const licenseKey = qData.license

  if (licenseKey.length !== LICENSE_KEY_LENGTH) {
    res.status(404)
    res.send({
      valid: false,
      error: { message: 'Invalid key.' }
    })
    return
  }

  if (context.provider.isOffline()) {
    res.status(400)
    res.send({
      error: {
        message: 'Bad request. Can not check license in offline mode.'
      }
    })
    return
  }

  validate(context, licenseKey, res)
}

function validate (context, licenseKey, res) {
  _validate(context, licenseKey, res, 0)
}

function _validate (context, licenseKey, res, tries) {
  if (tries >= 2) {
    res.status(500)
    res.send({
      time: new Date().toISOString(),
      valid: false,
      error: { message: 'Server error' }
    })
    Logger.error(`Too many errors validating ${licenseKey}, giving up`)
    return
  }

  context.provider.validateKey(licenseKey, (err, license) => {
    if (err) {
      if (err.status === 404) {
        res.status(404)
        res.send({
          time: new Date().toISOString(),
          valid: false,
          error: { message: err.message }
        })
        Logger.error(`Error checking license ${licenseKey}: ${err.message}`)
        return
      }

      // Some other error. Retry one time.
      Logger.error(`Error checking license ${licenseKey}, maybe retrying: ${err.message}`)
      setImmediate(() => {
        _validate(context, licenseKey, res, tries + 1)
      })
      return
    }

    if (!license) {
      res.status(404)
      res.send({
        time: new Date().toISOString(),
        valid: false
      })
      return
    }

    res.status(200)
    res.send({
      time: new Date().toISOString(),
      valid: license.valid,
      expired: (license.expires && license.expires < Date.now())
    })
  })
}
