'use strict'

const API = require('./API')
const queryExprs = require('../../query-expressions.js')

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

exports.initialize = initialize
exports.parse = parse

const VERSION = API.VERSION

const Parsers = {}
const Multiple = {}

// Initialize parsers from schema
// parser (val: string, errors: string[])
function initialize (context) {
  const schema = context.schemas.http[VERSION]
  const parms = schema.parm

  // validate we have all the parsers
  for (let parm in parms) {
    const multiple = parms[parm].multiple || false
    Multiple[parm] = multiple

    if (Parsers[parm] == null) {
      Logger.warn(`no parser defined for parm ${parm}`)
      Parsers[parm] = (context, val, errors) => {
        errors.push(`parsed ${parm} with default identity parser`)
      }
    }

    Parsers[parm].multiple = multiple
  }

  // fix any bits up that we need
  for (let method in schema) {
    if (['GET', 'POST', 'PUT', 'DELETE'].indexOf(method) === -1) continue
    for (let uri in schema[method]) {
      const apiSchema = schema[method][uri]
      if (apiSchema.parms == null) apiSchema.parms = []
    }
  }
}

// Parse the `req.query` object from the schema for the specified API

// Returns a new object with shape similar to req.query, but with parsed
// versions of the parms, and with properties errors and warnings with messages.
function parse (context, apiSchema, query) {
  Logger.debug(`parsing`, {apiSchema, query})

  const errors = []
  const data = {
    errors: errors
  }

  if (query == null) return data

  query = removeEmptyParams(query)

  // loop through query parameters
  for (let name in query) {
    let value = query[name]
    const parser = Parsers[name]

    if (parser == null) {
      errors.push(`unknown parameter ${name}`)
      continue
    }

    if (apiSchema.parms.indexOf(name) === -1) {
      errors.push(`unused parameter ${name}`)
      continue
    }

    if (Multiple[name] && !Array.isArray(value)) value = [ value ]
    if (!Multiple[name] && Array.isArray(value)) {
      errors.push(`parameter ${name} used multiple times`)
      continue
    }

    const parseErrors = []

    if (!Multiple[name]) {
      data[name] = parser(context, value, parseErrors)
    } else {
      data[name] = value
        .map(value => parser(context, value, parseErrors))
        .filter(value => value != null)
    }

    for (let message of parseErrors) {
      errors.push(`error parsing parameter ${name}: ${message}`)
    }
  }

  // check for required parms
  if (apiSchema.required == null) apiSchema.required = []

  for (let required of apiSchema.required) {
    if (data[required] != null) continue

    errors.push(`parameter required: ${required}`)
  }

  // ensure all parms with multiple values are empty arrays instead of null
  for (let parm of apiSchema.parms) {
    if (Multiple[parm] && data[parm] == null) data[parm] = []
  }

  return data
}

function removeEmptyParams (query) {
  var pruned = {}

  for (let key of Object.keys(query)) {
    if (query[key] == null) {
      continue
    }

    if (Array.isArray(query[key])) {
      let filtered = query[key].filter((elem) => {
        return (elem != null && elem !== '')
      })
      if (filtered.length !== 0) {
        pruned[key] = filtered
      }
    } else if (query[key] !== '') {
      pruned[key] = query[key]
    }
  }

  return pruned
}

Parsers.interval = parseDuration
Parsers.duration = parseDuration
Parsers.q = parseQueryExpr
Parsers.key = parseIdentity
Parsers.id = parseIdentity
Parsers.field = parseField
Parsers.start = parseTime
Parsers.end = parseTime
Parsers.name = parseIdentity
Parsers.data = parseIdentity
Parsers.license = parseIdentity
Parsers.token = parseIdentity
Parsers.type = parseIdentity
Parsers.app = parseIdentity
Parsers.hostname = parseIdentity
Parsers.tags = parseIdentity
Parsers.starred = parseBool
Parsers.visualization = parseIdentity
Parsers.series = parseRange(['raw', '1m', '1h'], 'raw')

function parseIdentity (context, value, errors) {
  return value
}

function parseBool (context, value, errors) {
  if (value === true || value === 'true' || value === '1') return true
  if (value === false || value === 'false' || value === '0') return false
  errors.push(`invalid boolean ${value}`)
}

function parseQueryExpr (context, value, errors) {
  let result
  try {
    result = queryExprs.parse(context, value)
  } catch (err) {
    errors.push(err.message)
  }

  return result
}

function parseField (context, value, errors) {
  const field = context.fieldManager.getField(value)
  if (field != null) return field

  errors.push(`ignoring unknown field ${value}`)
  return null
}

function parseTime (context, value, errors) {
  let match = value.match(/^\d{4}-.*/)
  const isISO = match != null

  let date

  if (isISO) {
    // try Date.parse()
    date = Date.parse(value)
    if (!isNaN(date)) return date
    errors.push(`invalid date format: ${value}`)
    return null
  }

  // try as a millsecond value > 1,000,000
  date = parseInt(value, 10)
  if (!isNaN(date)) {
    if (date >= 1000 * 1000) return date
  }

  // try as a duration
  let duration = parseDuration(context, value, [])
  if (duration != null) {
    return Date.now() - duration * 1000
  }

  // try Date.parse()
  date = Date.parse(value)
  if (!isNaN(date)) return date

  errors.push(`unable to parse date ${value}`)
  return null
}

// Regular expression for duration expressions.
const DurationRegex = /(-)?(\d+)(ms|s|m|h|d|w)?/

// return value in seconds units
function parseDuration (context, value, errors) {
  if (value == null) return null

  const match = value.match(DurationRegex)
  if (match == null) {
    errors.push('invalid duration expression')
    return null
  }

  let val = parseFloat(match[2])
  const units = match[3] || 's'

  switch (units) {
    case 'ms': return val / 1000
    case 's': return val
    case 'm': return val * 60
    case 'h': return val * 60 * 60
    case 'd': return val * 60 * 60 * 24
    case 'w': return val * 60 * 60 * 24 * 7
    default:
      errors.push(`invalid time unit ${units}`)
  }

  return null
}

function parseRange (options, defaultValue) {
  if (defaultValue === undefined) {
    throw new Error('Range parser requires a defined default value.')
  }
  function customParser (context, value, errors) {
    if (value == null) {
      return defaultValue
    }
    if (options.indexOf(value) >= 0) {
      return value
    }
    errors.push(`unrecognized value '${value}'. [${options.join(', ')}]`)
    return null
  }

  return customParser
}
