'use strict'

const util = require('../util')
const http = require('http')

exports.ping = ping
exports.write = write
exports.query = query
exports.shutDown = shutDown

let ShuttingDown = false

const Logger = util.getLogger(__dirname, __filename)

const clientRequest = require('client-request')
const encodeForm = require('form-urlencoded')
const encodeQuery = require('querystring').encode

const keepAliveAgent = new http.Agent({ keepAlive: true })

// Write a line to InfluxDB.
function write (url, db, data, rp, cb) {
  if (ShuttingDown) return setImmediate(cb)

  Logger.debug(`writing to ${db}`, data)

  const queryOpts = {
    db: db,
    precision: 'n' // nanoseconds; see: https://docs.influxdata.com/influxdb/v0.13/concepts/api/
  }

  // If using external influxdb and the db is blown away behind the console, it
  // will start producing errors like {"error":"retention policy not found:
  // agentData"} or {"error":"database not found: \"nsolid\""}.  It might be
  // desirable to refactor this code to recreate the schema if this is
  // detected.  For now, the server needs to be restarted if the database is
  // blown away.

  if (rp != null) queryOpts.rp = rp

  const qs = encodeQuery(queryOpts)
  const secure = (url.substr(0, 5) === 'https')

  clientRequest({
    uri: `${url}/write?${qs}`,
    agent: secure ? null : keepAliveAgent,
    method: 'POST',
    body: data
  }, onRequest)

  function onRequest (err, res, body) {
    if (ShuttingDown) return cb()

    if (err) {
      Logger.error(err, 'error writing data')
      Logger.debug(data)
      if (cb) {
        cb(err)
      }
      return
    }

    if (res.statusCode === 400) {
      Logger.error('syntax error writing data', body.toString())
      Logger.debug(data)
      if (cb) {
        cb(new Error('syntax error writing data'))
      }
      return
    }

    if (res.statusCode !== 204) {
      Logger.error('unknown error writing data', body.toString())
      Logger.debug(data)
      if (cb) {
        cb(new Error('unknown error writing data'))
      }
      return
    }

    if (cb) {
      cb()
    }
  }
}

// Run a query against InfluxDB.
function query (url, db, command, cb) {
  if (!cb) cb = err => Logger.error(`Error querying Influx: ${err.message}`)
  if (ShuttingDown) return setImmediate(cb)

  const timeStarted = Date.now()

  const form = encodeForm({
    db: db,
    epoch: 'ms',
    q: command
  })

  const secure = (url.substr(0, 5) === 'https')

  clientRequest({
    uri: `${url}/query`,
    agent: secure ? null : keepAliveAgent,
    method: 'POST',
    body: form,
    headers: {
      'content-type': 'application/x-www-form-urlencoded'
    }
  }, onRequest)

  function onRequest (err, res, body) {
    if (ShuttingDown) return cb()

    const timeElapsed = Date.now() - timeStarted
    const statusCode = res ? res.statusCode : 'n/a'
    const message = `query ${db}: ${command}; ${statusCode}; ${timeElapsed} ms`
    if (err) {
      Logger.error(err, message)
    } else {
      Logger.debug(message)
    }

    if (err) return cb(err)

    switch (res.statusCode) {
      case 200: return cb(null, body)
      case 204: return cb(null, body)
    }

    err = new Error('status ' + res.statusCode + '; command: ' + command)
    err.body = '' + body

    Logger.error(`error from query: ${body}`)
    cb(err)
  }
}

// Ping InfluxDB, return the version on success.
function ping (url, cb) {
  if (ShuttingDown) return setImmediate(cb)

  clientRequest({
    uri: `${url}/ping`,
    method: 'GET'
  }, onRequest)

  function onRequest (err, res, body) {
    if (ShuttingDown) return cb()
    if (err) return cb(err)

    if (res.statusCode !== 204) {
      const err = new Error(`ping status: ${res.statusCode}, body: ${body}`)
      return cb(err)
    }

    const version = res.headers['x-influxdb-version']
    if (version != null) return cb(null, version)

    err = new Error('ping did not return a X-Influxdb-Version header')
    return cb(err)
  }
}

function shutDown () {
  ShuttingDown = true
}
