'use strict'

exports.parse = parse
exports.fromJSON = fromJSON

const queryTerms = require('./query-terms')

// parse a filter expression
// returns QueryExpression object or throws error
function parse (context, string) {
  if (string == null || string === '') return create(context, [])

  // throws exception on bad parse
  const roughTerms = roughParse(string)

  return create(context, roughTerms)
}

// convert a JSON representation to a query expression
// returns QueryExpression object or throws error
// A QueryExpression  object can be converted to it's JSON form via
// `qe.toJSON()` or can be directly stringified via `JSON.stringify(qe)`
function fromJSON (context, object) {
  if (object == null) return create(context, [])
  if (!Array.isArray(object)) throw new Error('object parameter must be an array')

  return create(context, object)
}

// expects valid terms, throws exception if not valid
function create (context, termDataObjects) {
  const terms = termDataObjects.map(termData => queryTerms.create(context, termData))
  return new QueryExpression(terms)
}

// models a query expression
class QueryExpression {
  constructor (terms) {
    this.terms = terms
  }

  // return boolean of evaluating expression against agent
  evalForAgent (agent) {
    if (this.terms.length === 0) return true

    for (let term of this.terms) {
      if (!term.evalForAgent(agent)) return false
    }
    return true
  }

  // return a string representation
  toString () {
    return this.terms
      .map(term => term.toString())
      .join(' ')
  }

  toJSON () {
    return this.terms.map(term => {
      return {
        field: term.field.name,
        operator: term.operator,
        values: term.values
      }
    })
  }
}

// Parse a filter expression into array of {field:string, operator:string, values:string[]}.
// Handles escaping chars with a % escape character, but provides no semantic validation.
// Returns array of terms {field,operator,values} but can throw exception

// this splits an expression into the field and then everything else
const RegExpToOperator = /^(.+?)([=!<>~]+.*)$/

// this splits an expression without initial field into operator and then everything else
const RegExpOperator = /^([=!<>~]+)(.*)$/

function roughParse (string) {
  const result = []

  // trim incoming string; string will be "remainder to be parsed" for rest of fn
  string = `${string}`.trim()

  // while we still have stuff to parse ...
  while (string !== '') {
    const field = getField()
    if (field == null) throw new Error(`error parsing field at "${string.substr(10)}"`)

    const operator = getOperator()
    if (operator == null) throw new Error(`error parsing operator at "${string.substr(10)}"`)

    const values = getValues()
    if (values == null) throw new Error(`error parsing values at "${string.substr(10)}"`)

    result.push({
      field: field,
      operator: operator,
      values: values
    })
  }

  return result

  // get the field out of the existing string, and then remove from string
  function getField () {
    const match = string.match(RegExpToOperator)
    if (match == null) return null

    string = match[2]
    return match[1]
  }

  // get the operator out of the existing string, and then remove from string
  function getOperator () {
    const match = string.match(RegExpOperator)
    if (match == null) return null

    string = match[2]
    return match[1]
  }

  // get the values out of the existing string, then remove from string
  function getValues () {
    const result = []
    let roughValues = getRoughValues()

    // split by unescaped comma
    while (roughValues !== '') {
      // find the next unescaped comma
      const index = indexOfUnescaped(roughValues, ',')

      // if not, we're done!
      if (index === -1) {
        result.push(roughValues)
        roughValues = ''
        break
      }

      // split at the unescaped comma, saving first bit, processing the rest
      result.push(roughValues.substr(0, index))
      roughValues = roughValues.substr(index + 1)
    }

    // unescape result values
    return result.map(unescapeString)
  }

  // unescape a string's %X substrings
  function unescapeString (s) {
    const result = s.replace(/%(.)/g, '$1')
    return result
  }

  // get values not split on commas, skipping %,
  function getRoughValues () {
    const index = indexOfUnescaped(string, ' ')
    if (index === -1) {
      const result = string
      string = ''
      return result
    }

    const result = string.substr(0, index)
    string = string.substr(index + 1).trim()
    return result
  }

  // split at next matching search string excluding escaped search string
  function indexOfUnescaped (string, search) {
    let start = 0
    while (true) {
      let index = string.indexOf(search, start)
      if (index === -1) return -1
      if (index === 0) return 0
      if (string[index - 1] !== '%') return index
      start = index + 1
    }
  }
}

// a CLI tester that takes first arg, parses it, prints the JSON or error
if (require.main === module) cli()

function cli () {
  const queryString = require('querystring')
  const context = {
    fieldManager: require('./field-manager').create()
  }

  let result
  const asString = process.argv[2]
  try {
    result = parse(context, asString)
  } catch (err) {
    return console.log(err)
  }

  console.log('as string:')
  console.log(asString)
  console.log('')

  console.log('as query string:')
  console.log(queryString.stringify({q: asString}))
  console.log('')

  console.log('as JSON:')
  console.log(JSON.stringify(result, null, 4))
}
