'use strict'

exports.create = create

const queryTermEvaluator = require('./query-term-evaluator')

// Validate a term given object {field:string, operator:string, values:string[]}.
// Returns a QueryTerm object or throws exception for invalid terms
function create (context, termDataInput) {
  const field = termDataInput.field
  const operator = termDataInput.operator
  const values = termDataInput.values

  if (field == null) throw new Error('field is not defined')
  if (operator == null) throw new Error('operator is not defined')
  if (!Array.isArray(values)) throw new Error('values is not an array')

  const termData = validateTerm(context, termDataInput)

  return new QueryTerm(termData.field, termData.operator, termData.values)
}

// models a query term
class QueryTerm {
  constructor (field, operator, values) {
    this._field = field
    this._operator = operator
    this._values = values
  }

  get field () { return this._field }
  get operator () { return this._operator }
  get values () { return this._values }

  // return boolean of evaluating term against agent
  evalForAgent (agent) {
    const agentValue = this._field.valueFromAgent(agent)
    return queryTermEvaluator.evaluate(this, agentValue)
  }

  toString () {
    const values = this.values.map(value => {
      return `${value}`
        .replace('%', '%%')
        .replace(',', '%,')
        .replace(' ', '% ')
    })
    return `${this._field.name}${this._operator}${values.join(',')}`
  }
}

const ValidOps = {
  number: '< <= > >='.split(' '),
  string: '= != ~ !~'.split(' '),
  boolean: '= !='.split(' '),
  package: '= != < <= > >='.split(' ')
}
ValidOps.array = ValidOps.string

// Validates term data returning normalized objects
function validateTerm (context, termData) {
  const result = {}

  result.field = context.fieldManager.getField(termData.field)
  if (result.field == null) {
    throw new Error(`invalid field "${termData.field}"`)
  }

  // check operator
  result.operator = termData.operator
  const validOps = ValidOps[result.field.type]
  if (validOps == null) {
    throw new Error(`internal error: unknown field type for field "${termData.field}"`)
  }
  if (validOps.indexOf(termData.operator) === -1) {
    throw new Error(`invalid operator "${termData.operator}" for field "${termData.field}"`)
  }

  // check values
  result.values = []
  for (let value of termData.values) {
    if (result.field.name === 'id') {
      // agent id special case.  maybe this should be its own type and use coerce, but meh
      const agent = context.agentManager.byId(value)
      if (agent) {
        value = agent.id
      } else {
        // If we are matching an agent id and don't know about it, make it a wilcard
        value = `${value}*`
        result.operator = '~'
      }
    }
    const coercedValue = result.field.coerce(value)
    if (coercedValue == null) {
      throw new Error(`invalid value "${value}" for field "${termData.field}"`)
    }
    result.values.push(coercedValue)
  }

  // check cardinality of values
  if (result.values.length === 0) {
    throw new Error(`no value specified for field "${termData.field}", operator "${termData.operator}"`)
  }

  if (result.field.type !== 'string') {
    if (result.values.length > 1) {
      throw new Error(`only one value can be specified with field "${termData.field}", operator "${termData.operator}"`)
    }
  }

  return result
}
